use std::ffi::c_void;
use std::ffi::CString;
use std::mem;
use std::os::raw::c_char;

use vtables::VTable;
use vtables_derive::has_vtable;
use vtables_derive::virtual_index;
use vtables_derive::VTable;

use crate::CreateInterfaceFn;

pub type FileHandle = i32;

#[repr(i32)]
#[allow(dead_code)]
pub enum SeekType {
    Head,
    Current,
    Tail,
}

#[allow(dead_code)]
pub enum Open {
    RB,
    WB,
    A,
    ABPlus,
}

impl Into<*const c_char> for Open {
    fn into(self) -> *const c_char {
        match self {
            Open::RB => b"rb\0".as_ptr() as _,
            Open::WB => b"wb\0".as_ptr() as _,
            Open::A => b"a\0".as_ptr() as _,
            Open::ABPlus => b"ab+\0".as_ptr() as _,
        }
    }
}

#[has_vtable]
#[derive(VTable, Debug)]
pub struct IBaseFileSystem {}

#[allow(dead_code)]
impl IBaseFileSystem {
    #[virtual_index(0)]
    fn read_internal(&self, output: *mut u8, buffer_size: i32, file: FileHandle) -> i32 {}
    #[virtual_index(1)]
    fn write_internal(&self, input: *const u8, buffer_size: i32, file: FileHandle) -> i32 {}
    #[virtual_index(2)]
    fn open_internal(
        &self,
        file_name: *const c_char,
        options: *const c_char,
        path_id: *const c_char,
    ) -> FileHandle {
    }
    #[virtual_index(3)]
    pub fn close(&self, file: FileHandle) {}
    #[virtual_index(4)]
    pub fn seek(&self, file: FileHandle, offset: i32, seek_from: SeekType) {}
    #[virtual_index(5)]
    pub fn tell(&self, file: FileHandle) -> u32 {}
    #[virtual_index(6)]
    pub fn file_size(&self, file: FileHandle) -> usize {}
    #[virtual_index(7)]
    fn size_internal(&self, file_name: *const c_char, path_id: *const c_char) -> u32 {}
    #[virtual_index(8)]
    pub fn flush(&self, file: FileHandle) {}
    #[virtual_index(9)]
    fn precache_internal(&self, file_name: *const c_char, path_id: *const c_char) -> bool {}
    #[virtual_index(10)]
    fn file_exists_internal(&self, file_name: *const c_char, path_id: *const c_char) -> bool {}
    #[virtual_index(11)]
    fn is_file_writable_internal(&self, file_name: *const c_char, path_id: *const c_char) -> bool {}
    #[virtual_index(12)]
    fn set_file_writable_internal(
        &self,
        file_name: *const c_char,
        writable: bool,
        path_id: *const c_char,
    ) -> bool {
    }
    #[virtual_index(13)]
    fn get_file_time_internal(&self, file_name: *const c_char, path_id: *const c_char) -> u64 {}
    #[virtual_index(16)]
    fn unzip_file_internal(
        &self,
        file_name: *const c_char,
        path: *const c_char,
        destination: *const c_char,
    ) -> bool {
    }

    pub fn read(&self, output: &mut [u8], file: FileHandle) -> i32 {
        self.read_internal(output.as_mut_ptr(), output.len() as _, file)
    }
    pub fn write(&self, input: &[u8], file: FileHandle) -> i32 {
        self.write_internal(input.as_ptr(), input.len() as _, file)
    }
    pub fn open(&self, path: &str, options: Open, game_path: Option<&str>) -> Option<FileHandle> {
        let result = self.open_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            options.into(),
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        );
        if result != 0 {
            Some(result)
        } else {
            None
        }
    }
    pub fn size(&self, path: &str, game_path: Option<&str>) -> usize {
        self.size_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        ) as _
    }
    pub fn precache(&self, path: &str, game_path: Option<&str>) -> bool {
        self.precache_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        )
    }
    pub fn file_exsists(&self, path: &str, game_path: Option<&str>) -> bool {
        self.file_exists_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        )
    }
    pub fn is_file_writable(&self, path: &str, game_path: Option<&str>) -> bool {
        self.is_file_writable_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        )
    }
    pub fn set_file_writable(&self, path: &str, writable: bool, game_path: Option<&str>) -> bool {
        self.set_file_writable_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            writable,
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        )
    }
    pub fn get_file_time(&self, path: &str, game_path: Option<&str>) -> u64 {
        self.get_file_time_internal(
            CString::new(path).map_or(0 as _, |path| path.as_ptr()),
            game_path.map_or(0 as _, |path| {
                CString::new(path).map_or(0 as _, |path| path.as_ptr())
            }),
        )
    }

    pub fn from_ptr(ptr: *mut c_void) -> &'static IBaseFileSystem {
        unsafe { mem::transmute(ptr) }
    }

    pub fn from_factory(factory: CreateInterfaceFn) -> Result<&'static IBaseFileSystem, i32> {
        let mut result = 0;
        let ptr = unsafe { factory(b"VBaseFileSystem011\0".as_ptr() as _, &mut result) };
        if ptr.is_null() {
            Err(result)
        } else {
            Ok(IBaseFileSystem::from_ptr(ptr))
        }
    }
}
