use std::ffi::c_void;
use std::ffi::CStr;
use std::ffi::CString;
use std::fmt;
use std::mem;

use vtables::VTable;
use vtables_derive::has_vtable;
use vtables_derive::VTable;

use crate::CreateInterfaceFn;

#[has_vtable]
#[derive(VTable)]
pub struct INetworkStringTable {}

impl fmt::Debug for INetworkStringTable {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("INetworkStringTable")
            .field("table_id", &self.get_table_id())
            .field("name", &self.get_table_name())
            .field("num_strings", &self.get_num_strings())
            .field("max_strings", &self.get_max_strings())
            .field("entry_bits", &self.get_entry_bits())
            .finish()
    }
}

#[allow(dead_code)]
impl INetworkStringTable {
    pub fn get_table_name(&self) -> &CStr {
        unsafe { CStr::from_ptr(self.get_table_name_internal()) }
    }

    pub fn get_string(&self, string_number: i32) -> Option<&CStr> {
        let ptr = self.get_string_internal(string_number);
        if ptr.is_null() {
            None
        } else {
            unsafe { Some(CStr::from_ptr(ptr)) }
        }
    }

    pub fn get_string_userdata(&self, string_number: i32) -> Option<Vec<u8>> {
        let mut length = 0;
        let ptr = self.get_string_userdata_internal(string_number, &mut length) as *const u8;
        if ptr.is_null() {
            None
        } else {
            let mut vec = vec![0u8; length as _];
            for offset in 0..length as usize {
                vec[offset] = unsafe { *ptr.add(offset) };
            }
            Some(vec)
        }
    }

    pub fn set_string_userdata(&self, string_number: i32, data: &[u8]) -> Result<(), &'static str> {
        if data.len() > 2i32.pow(self.get_entry_bits() as _) as _ {
            Err("Data is too long")
        } else {
            self.set_string_userdata_internal(string_number, data.len() as _, data.as_ptr() as _);
            Ok(())
        }
    }

    pub fn from_ptr(ptr: *mut c_void) -> Option<&'static Self> {
        unsafe { (ptr as *mut Self).as_ref() }
    }
}

#[has_vtable]
#[derive(VTable, Debug)]
pub struct INetworkStringTableContainer {}

#[allow(dead_code)]
impl INetworkStringTableContainer {
    pub fn create_string_table(
        &self,
        table_name: &str,
        max_entries: i32,
        userdatafixedsize: Option<i32>,
        userdatanetworkbits: Option<i32>,
        is_file_names: Option<bool>,
    ) -> Option<&'static INetworkStringTable> {
        let cstr = CString::new(table_name).unwrap_or_default();
        let ptr = dbg!(self.create_string_table_ex_internal(
            cstr.as_ptr(),
            max_entries,
            userdatafixedsize.unwrap_or(0),
            userdatanetworkbits.unwrap_or(0),
            is_file_names.unwrap_or(false),
        ));
        dbg!(ptr);
        INetworkStringTable::from_ptr(ptr)
    }

    pub fn get_table(&self, string_table: i32) -> Option<&'static INetworkStringTable> {
        unsafe { self.get_table_internal(string_table).as_ref() }
    }

    pub fn find_table(&self, table_name: &str) -> Option<&'static INetworkStringTable> {
        unsafe {
            let cstr = CString::new(table_name).unwrap_or_default();
            self.find_table_internal(cstr.as_ptr()).as_ref()
        }
    }

    pub fn from_ptr(ptr: *mut c_void) -> &'static Self {
        unsafe { mem::transmute(ptr) }
    }

    pub fn from_factory(factory: CreateInterfaceFn) -> Result<&'static Self, i32> {
        let mut result = 0;
        let ptr = unsafe { factory(b"VEngineServerStringTable001\0".as_ptr() as _, &mut result) };
        if ptr.is_null() {
            Err(result)
        } else {
            Ok(Self::from_ptr(ptr))
        }
    }
}
