use std::ffi::{c_void, CStr};

use crate::{CreateInterfaceFn, PluginResult, QueryStatus, ServerPlugin};

#[allow(dead_code)]
pub struct BasePlugin<PLUGIN>
where
    PLUGIN: ServerPlugin,
{
    server_callbacks_vtable: *const *mut c_void,
    server_plugin: PLUGIN,
}

#[allow(unused_variables)]
impl<PLUGIN> BasePlugin<PLUGIN>
where
    PLUGIN: ServerPlugin,
{
    const PLUGIN_VTABLE: &'static [*mut c_void] = &[
        BasePlugin::<PLUGIN>::load as _,
        BasePlugin::<PLUGIN>::unload as _,
        BasePlugin::<PLUGIN>::pause as _,
        BasePlugin::<PLUGIN>::unpause as _,
        BasePlugin::<PLUGIN>::get_plugin_description as _,
        BasePlugin::<PLUGIN>::level_init as _,
        BasePlugin::<PLUGIN>::server_activate as _,
        BasePlugin::<PLUGIN>::game_frame as _,
        BasePlugin::<PLUGIN>::level_shutdown as _,
        BasePlugin::<PLUGIN>::client_active as _,
        BasePlugin::<PLUGIN>::client_disconnect as _,
        BasePlugin::<PLUGIN>::client_put_in_server as _,
        BasePlugin::<PLUGIN>::set_command_client as _,
        BasePlugin::<PLUGIN>::client_settings_changed as _,
        BasePlugin::<PLUGIN>::client_connect as _,
        BasePlugin::<PLUGIN>::client_command as _,
        BasePlugin::<PLUGIN>::network_id_validated as _,
        BasePlugin::<PLUGIN>::on_query_cvar_value_finished as _,
        BasePlugin::<PLUGIN>::on_edict_allocated as _,
        BasePlugin::<PLUGIN>::on_edict_freed as _,
    ];
    pub fn new(server_plugin: PLUGIN) -> Self {
        Self {
            server_callbacks_vtable: Self::PLUGIN_VTABLE.as_ptr(),
            server_plugin,
        }
    }
    #[inline(never)]
    unsafe fn load(
        &mut self,
        game_factory: CreateInterfaceFn,
        server_factory: CreateInterfaceFn,
    ) -> bool {
        match self.server_plugin.load(game_factory, server_factory) {
            Err(e) => false,
            _ => true,
        }
    }
    #[inline(never)]
    unsafe fn unload(&mut self) {
        self.server_plugin.unload()
    }
    #[inline(never)]
    unsafe fn pause(&mut self) {
        self.server_plugin.pause()
    }
    #[inline(never)]
    unsafe fn unpause(&mut self) {
        self.server_plugin.unpause()
    }
    #[inline(never)]
    unsafe fn get_plugin_description(&mut self) -> *const i8 {
        self.server_plugin.get_plugin_description().as_ptr()
    }
    #[inline(never)]
    unsafe fn level_init(&mut self, level_name: *const i8) {
        self.server_plugin.level_init(CStr::from_ptr(level_name))
    }
    #[inline(never)]
    unsafe fn server_activate(
        &mut self,
        edict_list: *mut c_void,
        edict_count: i32,
        max_clients: i32,
    ) {
        // TODO: Pass slice with acutal edicts.
        self.server_plugin.server_activate(&mut [], max_clients)
    }
    #[inline(never)]
    unsafe fn game_frame(&mut self, simulating: bool) {
        self.server_plugin.game_frame(simulating);
    }
    #[inline(never)]
    unsafe fn level_shutdown(&mut self) {
        self.server_plugin.level_shutdown();
    }
    #[inline(never)]
    unsafe fn client_active(&mut self, edict: *mut c_void) {
        self.server_plugin.client_active(edict);
    }
    #[inline(never)]
    unsafe fn client_disconnect(&mut self, edict: *mut c_void) {
        self.server_plugin.client_disconnect(edict);
    }
    #[inline(never)]
    unsafe fn client_put_in_server(&mut self, edict: *mut c_void, player_name: *const i8) {
        self.server_plugin
            .client_put_in_server(edict, CStr::from_ptr(player_name));
    }
    #[inline(never)]
    unsafe fn set_command_client(&mut self, client_index: i32) {
        self.server_plugin.set_command_client(client_index);
    }
    #[inline(never)]
    unsafe fn client_settings_changed(&mut self, edict: *mut c_void) {
        self.server_plugin.client_settings_changed(edict);
    }
    #[inline(never)]
    unsafe fn client_connect(
        &mut self,
        allow_connect: *mut bool,
        edict: *mut c_void,
        player_name: *const i8,
        player_address: *const i8,
        message_buffer: *mut i8,
        max_message_length: i32,
    ) -> PluginResult {
        match self.server_plugin.client_connect(
            edict,
            CStr::from_ptr(player_name),
            CStr::from_ptr(player_address),
        ) {
            Err(msg) => {
                *allow_connect = false;
                message_buffer.copy_from(msg.as_bytes().as_ptr() as _, max_message_length as _);
                PluginResult::Stop
            }
            Ok(res) => res,
        }
    }
    #[inline(never)]
    unsafe fn client_command(&mut self, edict: *mut c_void, args: *const c_void) -> PluginResult {
        self.server_plugin.client_command(edict, args)
    }
    #[inline(never)]
    unsafe fn network_id_validated(
        &mut self,
        player_name: *const i8,
        network_id: *const i8,
    ) -> PluginResult {
        self.server_plugin
            .network_id_validated(CStr::from_ptr(player_name), CStr::from_ptr(network_id))
    }
    #[inline(never)]
    unsafe fn on_query_cvar_value_finished(
        &mut self,
        cookie: i32,
        edict: *mut c_void,
        status: QueryStatus,
        cvar_name: *const i8,
        cvar_value: *const i8,
    ) {
    }
    #[inline(never)]
    unsafe fn on_edict_allocated(&mut self, edict: *mut c_void) {}
    #[inline(never)]
    unsafe fn on_edict_freed(&mut self, edict: *const c_void) {}
}
